use super::error::{ParserError, ParserErrorKind};
use crate::{
    lexer::{Symbol, Token, TokenKind},
    span::Span,
};

/// The type of a node in the AST.
#[derive(Debug, Clone, PartialEq)]
pub enum NodeKind {
    /// Bool.
    Bool(bool),
    /// Character.
    Char(char),
    /// Keyword.
    Keyword(Symbol),
    /// Number.
    Number(f64),
    /// String.
    String(String),
    /// Symbol.
    Symbol(Symbol),
    /// Nil.
    Nil,

    /// List.
    List(Vec<Node>),
    /// Vector.
    Vector(Vec<Node>),
}

impl std::fmt::Display for NodeKind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            NodeKind::Bool(_) => write!(f, "BOOL"),
            NodeKind::Char(_) => write!(f, "CHAR"),
            NodeKind::Keyword(_) => write!(f, "KEYWORD"),
            NodeKind::Number(_) => write!(f, "NUMBER"),
            NodeKind::String(_) => write!(f, "STRING"),
            NodeKind::Symbol(_) => write!(f, "SYMBOL"),
            NodeKind::Nil => write!(f, "NIL"),
            NodeKind::List(_) => write!(f, "LIST"),
            NodeKind::Vector(_) => write!(f, "VECTOR"),
        }
    }
}

/// A node in the AST with a start and end location.
#[derive(Debug, Clone, PartialEq)]
pub struct Node {
    /// The type of node.
    pub kind: NodeKind,
    /// The span in which the node occurs.
    pub span: Span,
}

impl Node {
    /// Construct a new instance of `Node`.
    #[must_use]
    pub const fn new(kind: NodeKind, span: Span) -> Self {
        Self { kind, span }
    }

    /// Push a child node onto a list node.
    pub fn push_node(&mut self, child: Self) -> Result<(), ParserError> {
        match &mut self.kind {
            NodeKind::List(c) | NodeKind::Vector(c) => {
                c.push(child);
            }
            _ => return Err(ParserError::new(ParserErrorKind::Unreachable, child.span)),
        }

        Ok(())
    }

    fn display(&self, indent: usize) -> String {
        let mut text = format!(
            "{}{}@{}..{}\n",
            "  ".repeat(indent),
            self.kind,
            self.span.bytes().start,
            self.span.bytes().end
        );

        match &self.kind {
            NodeKind::List(vec) | NodeKind::Vector(vec) => {
                for node in vec {
                    text.push_str(&node.display(indent + 1));
                }
            }
            _ => {}
        }

        text.trim_end().to_string()
    }
}

impl std::fmt::Display for Node {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.display(0))
    }
}

impl TryFrom<Token> for Node {
    type Error = ParserError;

    fn try_from(token: Token) -> Result<Self, Self::Error> {
        let kind = match token.kind {
            TokenKind::Bool(b) => NodeKind::Bool(b),
            TokenKind::Char(c) => NodeKind::Char(c),
            TokenKind::Number(n) => NodeKind::Number(n),
            TokenKind::String(s) => NodeKind::String(s),
            TokenKind::Keyword(k) => NodeKind::Keyword(k),
            TokenKind::Symbol(s) => NodeKind::Symbol(s),
            TokenKind::Nil => NodeKind::Nil,
            TokenKind::OpenParen
            | TokenKind::CloseParen
            | TokenKind::OpenBrace
            | TokenKind::CloseBrace
            | TokenKind::OpenBracket
            | TokenKind::CloseBracket
            | TokenKind::LineComment(_)
            | TokenKind::BlockComment(_) => {
                return Err(ParserError::new(ParserErrorKind::Unreachable, token.span))
            }
        };

        Ok(Self::new(kind, token.span))
    }
}
